import numpy as np

class BaseModel:

    max_gen_default = 1024
    model_max = 2048

    def __init__(self, model_name, global_conter=None, model_type = "chat"):
        self.name = model_name
        assert model_type in ["chat", "completion"], "type should be either 'chat' or 'completion'"
        self.type = model_type
        self.global_counter = global_conter
        self.debug = False
    
    def query_chat(self, messages, shrink_idx, model, max_gen=1024, temp=0.):
        raise NotImplementedError
    
    def query_completion(self, messages, shrink_idx, model, max_gen=1024, temp=0.):
        raise NotImplementedError
    
    def encode(self, txt):
        print("Warning: encode is not implemented for this model, returning words.")
        return txt.split(" ")

    def decode(self, txt):
        print("Warning: decode is not implemented for this model, returning concatenation of input.")
        return " ".join(txt)
    
    def count_tokens(self, text):
        if type(text) == list:
            return min(sum([len(self.encode(x['content'])) for x in text]), self.model_max)
        elif type(text) == str:
            return min(len(self.encode(text)), self.model_max)
        else:
            raise ValueError("Unknown type for text: {}".format(type(text)))

    def __call__(self, msg, shrink_idx, max_gen=None, temp=0.):
        if max_gen is None:
            max_gen = self.max_gen_default

        assert type(msg) == list, "msg should be a list in openai format"
        
        if self.type == "chat":
            result, usage = self.query_chat(msg, shrink_idx, max_gen, temp)
        else:
            result, usage = self.query_completion(msg, shrink_idx, max_gen, temp)
        
        if self.global_counter is not None:
            self.global_counter["token_completion"][self.name] += usage["completion"]
            self.global_counter["token_prompt"][self.name] += usage["prompt"]
            self.global_counter["api_calls"][self.name] += 1
        
        return result, usage

    

    def shrink_msg_by(self, msg, shrink_idx, L):
        if L <= 0:
            return msg
        counter = 0
        while L > 0:
            L_old = len(self.encode(self.compile_msg_txt(msg)))
            print("Shrinking message by {} tokens...".format(L))
            if len(self.encode(msg[shrink_idx]['content']))<L:
                shrink_idx = np.argmax([len(self.encode(x['content'])) for x in msg])
            msg[shrink_idx]['content'] = self.decode(self.encode(msg[shrink_idx]['content'])[L:])
            L_new = len(self.encode(self.compile_msg_txt(msg)))
            L = L - (L_old - L_new)
            if L_old - L_new <= 0:
                counter+=1
                if counter>=3:
                    print("Shrinking failed!")
                    import ipdb; ipdb.set_trace()
            else:
                counter = 0
        return msg
    
    def compile_msg_txt(self, msg):
        result = ""
        for x in msg:
            if x['role'].lower() == 'system':
                role = "context"
            elif x['role'].lower() == 'user':
                role = "question"
            else:
                role = "answer"
            result += "## {} ##\n".format(role)
            result += x['content'].strip() + "\n\n"
        result += "## answer ##\n"
        return result.strip()

    def compute_length(self, msg):
        L = 0
        for x in msg:
            L += len(self.encode(x['content']))
        return L

    def shrink_msg(self, msg, shrink_idx, L_max):
        L = self.compute_length(msg)
        return self.shrink_msg_by(msg, shrink_idx, L-L_max)

    def shrink_text(self, msg, shrink_idx, L_max):
        L = len(self.encode(self.compile_msg(msg)))
        return self.shrink_msg_by(msg, shrink_idx, L-L_max)
    
    def shrink_raw(self, text, L_max):
        L = len(self.encode(text))
        if L>L_max:
            return self.encode(text)[L-L_max+1:]
        else:
            return text


import json
def extract_json_objects(input_string):
    try:
        # Initialize variables
        json_objects = []
        stack = []
        start_index = -1
        
        # Helper function to try to decode a JSON object string
        def try_decode_json(json_str):
            try:
                # Attempt to parse the string as JSON
                return json.loads(json_str)
            except json.JSONDecodeError:
                # If an error occurs, return None
                return None

        # Iterate over each character and its index in the input string
        for index, char in enumerate(input_string):
            # Check for opening braces/brackets
            if char in '{[':
                # If this is the first opening brace/bracket, take note of the index
                if not stack:
                    start_index = index
                # Push the character onto the stack
                stack.append(char)
            
            # Check for closing braces/brackets
            elif char in '}]':
                if not stack:
                    continue  # Ignore unmatched closing braces/brackets

                # Check if the closing brace/bracket matches the last opening one
                if (char == '}' and stack[-1] == '{') or (char == ']' and stack[-1] == '['):
                    stack.pop()  # Pop from stack for a valid closing match
                    # When stack is empty, we have a full JSON string
                    if not stack:
                        # Extract the JSON string and try to decode it
                        json_str = input_string[start_index:index+1]
                        json_obj = try_decode_json(json_str)
                        # If successfully decoded, add to the json_objects list
                        if json_obj is not None:
                            json_objects.append(json_obj)
                else:
                    continue  # Skip unmatched closing brace/bracket
        if len(json_objects)==0:
            return None, "Error: No json objects found"
        return json_objects, None
    except Exception as e:
        return None, "Error: {}\nTraceback: {}".format(e, traceback.format_exc())


token_counter = {
    "token_completion":{},
    "token_prompt":{},
    "api_calls":{}
}

def get_token_counts():
    global token_counter
    return token_counter

def get_query(LLM_name):
    global token_counter
    token_counter["token_completion"][LLM_name] = 0
    token_counter["token_prompt"][LLM_name] = 0
    token_counter["api_calls"][LLM_name] = 0

def query_gpt_chat(model):
    from .GPT import GPT_chat
    global token_counter
    query_model = GPT_chat(model, token_counter)
    return query_model

def query_claude_chat(model):
    from .claude import Claude_chat
    global token_counter
    query_model = Claude_chat(model, token_counter)
    return query_model

def get_query(LLM_name):
    """Get the query model for the specified LLM_name.

    Currently supported LLMs:

    - GPT-4

    - GPT-3.5

    - Claude
    
    Args:
        LLM_name (str): Name of the LLM model
    
    Returns:
        LLM_API_FUNCTION: The query model for the specified LLM_name
    """
    global token_counter
    token_counter["token_completion"][LLM_name] = 0
    token_counter["token_prompt"][LLM_name] = 0
    token_counter["api_calls"][LLM_name] = 0

    if LLM_name.lower().startswith("gpt-4") or LLM_name.lower().startswith("gpt-3.5"):
        return query_gpt_chat(model=LLM_name)
    elif LLM_name.lower().startswith("claude"):
        return query_claude_chat(model=LLM_name)
    else:
        raise NotImplementedError("LLM {} not implemented".format(LLM_name))


from collections import deque
import datetime
try:
    from wandb.sdk.data_types.trace_tree import Trace
except:
    print("wandb Trace seems to be missing. This is fine if you are not using wandb")
    pass

class Graph:
    """A class to represent a DAG.

    This class represents a DAG with nodes and edges. It is used to represent
    the dependencies between different nodes in a graph. The graph can be
    evaluated in a topological order. The graph can also have temporary nodes
    and edges. Temporary nodes and edges are used to represent dynamic changes
    in the graph during evaluation. Temporary nodes and edges are cleared after
    each evaluation. The evaluation processs can also be logged to wandb.

    Attributes:
        nodes (dict): A dictionary of all nodes in the graph.
        temporary_nodes (dict): A dictionary of all temporary nodes in the graph.
        temporary_edges (list): A list of all temporary edges in the graph.
        temporary_removed_edges (list): A list of all edges that have been temporarily removed.
        history (dict): A dictionary of all results from the graph.
        num_iter (int): The number of iterations the graph has gone through.
        history_list (list): A list of all results from the graph.
        order (list): A list of the order in which the nodes were evaluated.
        queue (list): A list of nodes that need to be evaluated.
        wandb_root_span (wandb.sdk.data_types.trace_tree.Trace): The root span for wandb logging.
        chain_span (wandb.sdk.data_types.trace_tree.Trace): The chain span for logging the current step.
    """
    def __init__(self):
        self.nodes = {}  # Dictionary to store all nodes
        self.temporary_nodes = {}
        self.temporary_edges = []
        self.temporary_removed_edges = []
        self.history = {}  # Dictionary to store all results
        self.num_iter = 0
        self.history_list = []
        self.order = []
        self.queue = []
        self.wandb_root_span = None
        self.chain_span = None
    
    def get_node_with_temporary(self, key):
        """Get a node from the graph.

        Args:
            key (str): The key of the node to get.
        
        Returns:
            Node: The node with the key.

        Raises:
            ValueError: If the key is not found in the graph.
        """

        if key in self.nodes.keys():
            return self.nodes[key]
        elif key in self.temporary_nodes.keys():
            return self.temporary_nodes[key]
        else:
            raise ValueError("Node ({}) not found in graph".format(key))

    def add_node(self, node):
        """Add a node to the graph.

        Args:
            node (Node): The node to add to the graph.
        
        Raises:
            AssertionError: If the node already exists in the graph.
        """
        assert node.key not in self.nodes.keys(), "Node ({}) already exists".format(node.key)
        self.nodes[node.key] = node

    def add_temporary_node(self, node):
        """Add a temporary node to the graph.

        Args:
            node (Node): The temporary node to add to the graph.
        
        Raises:
            AssertionError: If the node already exists in the graph.
        """
        assert node.key not in self.nodes.keys(), "Node ({}) already exists in permanent graph".format(node.key)
        assert node.key not in self.temporary_nodes.keys(), "Node ({}) already exists in temporary graph".format(node.key)
        self.temporary_nodes[node.key] = node
        
    def has_edge_with_temporary(self, from_key, to_key):
        """Check if there is an edge between two nodes in the graph.

        Args:
            from_key (str): The key of the node to check from.
            to_key (str): The key of the node to check to.
        
        Returns:
            bool: True if there is an edge between the two nodes, False otherwise.
        
        Raises:
            AssertionError: If the from_key is not found in the graph.
        """
        node_from = self.get_node_with_temporary(from_key)
        node_to = self.get_node_with_temporary(to_key)
        assert node_from is not None and node_to is not None, "Node ({}) not found in graph".format(from_key)
        return node_from in node_to.adjacent_from

    def add_edge(self, from_key, to_key, prepend=False):
        """Add an edge between two nodes in the graph.

        Args:
            from_key (str): The key of the node to add the edge from.
            to_key (str): The key of the node to add the edge to.
            prepend (bool): Whether to prepend the edge to the front of the adjacent_from list.
        
        Raises:
            ValueError: If the from_key or to_key is not found in the graph.
        """
        if from_key in self.nodes.keys() and to_key in self.nodes.keys():
            assert not self.has_edge_with_temporary(from_key, to_key), "Edge ({}) already exists".format((from_key, to_key))
            self.nodes[from_key].adjacent_to.append(self.nodes[to_key])
            if prepend:
                self.nodes[to_key].adjacent_from.insert(0, self.nodes[from_key])
            else:
                self.nodes[to_key].adjacent_from.append(self.nodes[from_key])
        else:
            raise ValueError("Node ({}, {}) not found in graph".format(from_key, to_key))
    
    def add_order(self, from_key, to_key):
        """Add an order (non-dependency edge) between two nodes in the graph.

        This function can be used to specify the order without adding an edge, 
        in order to avoid a node being evaluated before another node.


        Args:
            from_key (str): The key of the node to add the order from.
            to_key (str): The key of the node to add the order to.
        
        Raises:
            AssertionError: If the from_key or to_key is not found in the graph.
            AssertionError: If the edge already exists between the two nodes.
        
        Note:
            This function does not add an edge between the two nodes.
        """
        assert from_key in self.nodes.keys() and to_key in self.nodes.keys(), "Node not found in graph"
        assert not self.has_edge_with_temporary(from_key, to_key), "Edge {} already exists. No need to specify order".format((from_key, to_key))
        self.nodes[to_key].evaluate_after.append(self.nodes[from_key])

    def add_edge_temporary(self, from_key, to_key, prepend=False):
        """Add a temporary edge between two nodes in the graph.

        This function adds a temporary edge between the two nodes. Temporary edges
        are used to represent dynamic changes in the graph during evaluation. Temporary
        edges are cleared after each evaluation.

        Args:
            from_key (str): The key of the node to add the edge from.
            to_key (str): The key of the node to add the edge to.
            prepend (bool): Whether to prepend the edge to the front of the adjacent_from list.
        
        Raises:
            AssertionError: If the from_key or to_key is not found in the graph.
            AssertionError: If the edge already exists between the two nodes.
        
        Note:
            It is recommended to use temporary edges to represent only dynamic changes in the graph.
        """
        node_from = self.get_node_with_temporary(from_key)
        node_to = self.get_node_with_temporary(to_key)
        assert node_from is not None and node_to is not None, "Node ({}) not found in graph".format(from_key)
        assert not self.has_edge_with_temporary(from_key, to_key), "Edge ({}) already exists".format((from_key, to_key))
        node_from.adjacent_to.append(node_to)
        if prepend:
            node_to.adjacent_from.insert(0, node_from)
        else:
            node_to.adjacent_from.append(node_from)
        if (from_key, to_key) in self.temporary_removed_edges:
            self.temporary_removed_edges.remove((from_key, to_key))
        else:
            self.temporary_edges.append((from_key, to_key))

        assert to_key not in self.history.keys(), "Cannot add edge to a node ({}) that has already been evaluated".format(to_key)
        if from_key in self.history.keys() and to_key not in self.queue: # if from_key has been evaluated, add to queue
            self.queue.append(to_key)
        

    def remove_edge_temporary(self, from_key, to_key):
        """Remove a temporary edge between two nodes in the graph.

        This function temporarily removes a edge between the two nodes. Temporary
        removals are reverted after each evaluation.

        Args:
            from_key (str): The key of the node to remove the edge from.
            to_key (str): The key of the node to remove the edge to.

        Raises:
            AssertionError: If the edge does not exist between the two nodes.
            AssertionError: If the from_key or to_key is not found in the graph.
        
        Note:
            It is recommended to use this function only for dynamic changes in the graph.
        """
        assert self.has_edge_with_temporary(from_key, to_key), "Edge does not exist"
        node_from = self.get_node_with_temporary(from_key)
        node_to = self.get_node_with_temporary(to_key)
        node_from.adjacent_to.remove(node_to)
        node_to.adjacent_from.remove(node_from)
        if (from_key, to_key) in self.temporary_edges:
            self.temporary_edges.remove((from_key, to_key))
        else:
            self.temporary_removed_edges.append((from_key, to_key))

        assert to_key not in self.history.keys(), "Cannot remove edge to a node ({}) that has already been evaluated".format(to_key)
        assert from_key not in self.history.keys(), "Cannot remove edge from a node ({}) that has already been evaluated".format(from_key)

    def skip_nodes_temporary(self, keys):
        """Skip nodes temporarily in the graph.

        This function temporarily skips nodes in the graph.
        The previously recorded outputs of the skipped nodes
        will be reused in the evaluation. Temporary skips are
        reverted after each evaluation.

        Args:
            keys (list): A list of keys of the nodes to skip.
        
        Raises:
            AssertionError: If the key is not found in the graph.
        
        Note:
            It is recommended to use this function only for dynamic changes in the graph.
        """
        for key in keys:
            node = self.get_node_with_temporary(key)
            # assert node is not temporary
            assert key not in self.temporary_nodes.keys(), "Cannot skip temporary node: {}".format(key)
            assert key not in self.history.keys(), "Cannot skip node {}. It has already been evaluated.".format(key)
            node.skip_turn()

    def clean_temporary(self):
        for from_key, to_key in self.temporary_edges:
            node_from = self.get_node_with_temporary(from_key)
            node_to = self.get_node_with_temporary(to_key)
            node_from.adjacent_to.remove(node_to)
            node_to.adjacent_from.remove(node_from)
        for from_key, to_key in self.temporary_removed_edges:
            node_from = self.get_node_with_temporary(from_key)
            node_to = self.get_node_with_temporary(to_key)
            node_from.adjacent_to.append(node_to)
            node_to.adjacent_from.append(node_from)
        self.temporary_nodes = {}
        self.temporary_edges = []
        self.temporary_removed_edges = []
    
    def set_wandb_root_span(self, span):
        self.wandb_root_span = span
    
    def set_trace(self):
        if self.wandb_root_span is None:
            return
        chain_span = Trace(
            "GraphChain",
            kind="chain",
            start_time_ms=round(datetime.datetime.now().timestamp() * 1000),
            metadata={"num_iter": self.num_iter},
        )
        self.chain_span = chain_span

    def commit_trace(self):
        if self.chain_span is None:
            return
        self.chain_span._span.end_time_ms = round(datetime.datetime.now().timestamp() * 1000)
        self.chain_span.add_inputs_and_outputs(inputs={}, 
                                               outputs={"response": self.history})
        self.wandb_root_span.add_child(self.chain_span)
        self.chain_span = None
    
    def get_streaming_history(self):
        assert len(self.history) == 0, "Error: This function should only be called before evaluate() is called."
        return self.history

    def evaluate(self):
        """Evaluate the graph in a topological order.

        This function evaluates the graph in a topological order. The order of evaluation
        is determined by the dependencies between the nodes. The graph can also have temporary
        nodes and edges. Temporary nodes and edges are used to represent dynamic changes in the
        graph during evaluation. Temporary nodes and edges are cleared after each evaluation.

        Returns:
            dict: A dictionary of the results from the graph.

        Raises:
            AssertionError: If the temporary nodes are not cleared before calling evaluate().
        """
        
        def recalculate_in_degree(key):
            node = self.get_node_with_temporary(key)
            return len([n for n in node.get_dependencies_inc_order() if n.key not in self.history.keys()]) 
        
        def find_next_node(queue):
            for key in queue:
                degree = recalculate_in_degree(key)
                if degree == 0:
                    return key
            return None
        
        assert len(self.temporary_nodes) == 0, "Temporary nodes must be cleared before calling evaluate()"
        in_degree = {key: len(self.nodes[key].adjacent_from) for key in self.nodes.keys()}
        self.queue = [key for key in in_degree.keys() if in_degree[key] == 0]
        self.order = []
        self.set_trace()
        while True:
            node_key = find_next_node(self.queue)
            if node_key is None:
                break
            self.queue.remove(node_key)
            self.order.append(node_key)
            node = self.get_node_with_temporary(node_key)
            self.history[node_key] = node.evaluate() # this may change the graph
            for adjacent in node.adjacent_to:
                if adjacent.key not in self.history.keys() and adjacent.key not in self.queue:
                    self.queue.append(adjacent.key)
        self.num_iter += 1
        self.commit_trace()
        self.history_list.append(self.history.copy())
        self.history = {}
        self.order = []
        self.queue = []
        self.clean_temporary()
        return self.history_list[-1]